/* chiffre.h : paramètres de chiffrement */

#ifndef CHIFFRE_H
#define CHIFFRE_H

/* *********
 * OPTIONS *
 * *********/

#define DEBUG   // Pour afficher notamment la table de chiffrement

/* Options pour choisir la méthode de chiffrement */
// #define XOR_SIMPLE
// #define XOR_FICHIER
#define PLAYFAIR

/* ******** */


#ifdef XOR_SIMPLE
/* Fonction de chiffrement avec xOR, en utilisant une clé fixe de 64 bits. */
extern void chiffre_xor_simple(char *buf, int l);
#endif

#ifdef XOR_FICHIER
#define LCLE 8192   // longueur de la clé : 8ko = 65536 bits, ATTENTION, doit être un multiple de 8*n
/* Fonction qui initialise la clé à partir d'un fichier externe. */
extern int init_cle_xor(void);
/* Fonction de chiffrement avec xOR, en utilisant une clé externe provenant d'un fichier. */
extern void chiffre_xor_fichier(char *buf, int l);
#endif

#ifdef PLAYFAIR
#define CLE_PLAYFAIR "e2l"
#define SAUT 29 // valeur du saut pour remplir le reste de la table
#define NBO 256 // nb d'octets
#define LCM 6   // Longueur du côté de la matrice

/* Fonction qui initialise la table de chiffrement. */
extern void init_table_chiffre(char *cle);
/* Fonction qui chiffre directement dans le buffer des données à chiffrer */
extern void chiffre_playfair(char *B, int T); // B adresse du buf, T le nb d'octets
/* Fonction inverse de chiffre_playfair, puisque cette fonction n'est pas involutive */
extern void dechiffre_playfair(char *B, int T); // B adresse du buf, T le nb d'octets
#endif
#endif
